package org.planet.snap;

import dermi.*;
import org.planet.p2pcm.*;
import org.planet.snap.interception.*;
import dermi.util.*;
import dermi.registry.*;
import dermi.bittorrent.*;
import dermi.exception.*;

import org.apache.axis.client.AdminClient;

import java.io.*;
import java.net.*;
import java.sql.*;
import java.util.*;
import java.util.zip.*;
import java.util.logging.*;

import javax.naming.*;
import java.security.PublicKey;
import java.security.Signature;

import org.hsqldb.Server;
import org.hsqldb.replication.Membership;
import org.hsqldb.replication.MembershipData;
import org.hsqldb.replication.MembershipImpl;
import org.hsqldb.replication.MembershipFactory;

public class Application extends DermiRemoteObject implements IApplication {
  private static Logger logger = Logger.getLogger (Application.class.getName());
  private String serverName = null;
  private int serverPort;
  private Properties appProperties = null;
  private DHTHandler storageContext = null;
  private BittorrentHandler btHandler = null;
  private static int firstUsableDBPort;
  private static ArrayList usedDBPorts = new ArrayList();
  private boolean dbException = false;
  private static boolean databaseBound = false;
  private static long lastDatabaseCall;
  private static Server s = null;
  private InitialContext context = null;
  private ComponentFactory memberFactory = null;

  // Default database passivation timeout (in minutes)
  private static int dbPassivationThreshold = 30;

  private Membership membership = null;

  // Dermi constructor: REQUIRED
  public Application() {
  }

  // Dermi constructor: REQUIRED
  public Application (Properties env) throws RemoteException {
    super (env);
    System.out.println ("[SnapApplication::snap] initialized.");
  }

  public Application (String serverName, Properties appConfig) throws RemoteException {
    super (appConfig);
    this.serverName = serverName;
    appProperties = appConfig;

    try {
      context = new InitialContext ();
    } catch (NamingException e) {
      e.printStackTrace();
      throw new RemoteException ("Unable to get JNDI tree reference.");
    }

    if (appProperties.getProperty (Context.DBINITIALPORT) != null && !databaseBound) {
      try {
        firstUsableDBPort = Integer.parseInt (appProperties.getProperty (Context.DBINITIALPORT));
        dbPassivationThreshold = Integer.parseInt (appProperties.getProperty (Context.DBPASSTHRESH));
      } catch (NumberFormatException nb) {
        throw new RemoteException ("Exception parsing '" + Context.DBINITIALPORT + "' property in snap-war.xml descriptor file.");
      }
    }

    // Read server port info (if available)
    try {
      DataInputStream d = new DataInputStream (new FileInputStream ("./.keepme"));
      this.serverPort = d.readInt();
    } catch (Exception e) { e.printStackTrace(); }


    storageContext = new DHTHandler (appProperties.getProperty (Context.BUNSHIN_PROPS), appProperties.getProperty (Context.SNAP_APPCONTEXT), appProperties.getProperty (Context.SNAP_CODEBASE) + File.separator + "storage", "http://" + serverName + ":" + serverPort + "/" + appProperties.getProperty (Context.SNAP_APPCONTEXT) + "/storage", null);
    //btHandler = SingletonHandler.getInstance("dermi.bittorrent.azureus.AzureusHandler");

    System.out.println ("[SnapApplication::" + appProperties.getProperty (Context.SNAP_APPCONTEXT) + "] initialized.");
	  
//    if ((appProperties.getProperty (Context.SNAP_PERSISTENCE) == null) ||
//        (!appProperties.getProperty (Context.SNAP_PERSISTENCE).equals (Context.DATABASE))) {

     // Clustering management is done without database mgmt
     String groupName = appProperties.getProperty (Context.SNAP_APPURL);
     if (!groupName.equals ("p2p://snap.urv.net")) {
       try {
         addWebAppMember();
       } catch (Exception e) {
         e.printStackTrace();
         throw new RemoteException ("Unable to add web application instance to group.");
       }
       int maxClusterSize = (appProperties.getProperty (Context.CLUSTERING) != null) ?
           Integer.parseInt (appProperties.getProperty (Context.CLUSTERING)) : 0;
       if (maxClusterSize > 0) {
         // Start clusteringThread (will form the application cluster, and keep it alive)
         // Check for correct cluster size each minute
         Timer t = new Timer();
         t.scheduleAtFixedRate (new ClusteringTimer (groupName, maxClusterSize, maxClusterSize, false), 20000, 60000);
       }
     }
	  deploySnapWebServices("http://" + serverName + ":" + serverPort + "/" + appProperties.getProperty (Context.SNAP_APPCONTEXT), appProperties);
 }

  public Application (String serverName, int serverPort, Properties appConfig) throws RemoteException {
    super (appConfig);
    this.serverName = serverName;
    this.serverPort = serverPort;
    appProperties = appConfig;

    try {
      context = new InitialContext ();
    } catch (NamingException e) {
      e.printStackTrace();
      throw new RemoteException ("Unable to get JNDI tree reference.");
    }

    if (appProperties.getProperty (Context.DBINITIALPORT) != null && !databaseBound) {
      try {
        firstUsableDBPort = Integer.parseInt (appProperties.getProperty (Context.DBINITIALPORT));
        dbPassivationThreshold = Integer.parseInt (appProperties.getProperty (Context.DBPASSTHRESH));
      } catch (NumberFormatException nb) {
        throw new RemoteException ("Exception parsing '" + Context.DBINITIALPORT + "' property in snap-war.xml descriptor file.");
      }
    }

    storageContext = new DHTHandler (appProperties.getProperty (Context.BUNSHIN_PROPS), appProperties.getProperty (Context.SNAP_APPCONTEXT), appProperties.getProperty (Context.SNAP_CODEBASE) + File.separator + "storage", "http://" + serverName + ":" + serverPort + "/" + appProperties.getProperty (Context.SNAP_APPCONTEXT) + "/storage", null);
    //btHandler = SingletonHandler.getInstance("dermi.bittorrent.azureus.AzureusHandler");

    System.out.println ("[SnapApplication::" + appProperties.getProperty (Context.SNAP_APPCONTEXT) + "] initialized.");

    

//    if ((appProperties.getProperty (Context.SNAP_PERSISTENCE) == null) ||
//        (!appProperties.getProperty (Context.SNAP_PERSISTENCE).equals (Context.DATABASE))) {

     // Clustering management is done without database mgmt
     String groupName = appProperties.getProperty (Context.SNAP_APPURL);
     if (!groupName.equals ("p2p://snap.urv.net")) {
       try {
         addWebAppMember();
       } catch (Exception e) {
         e.printStackTrace();
         throw new RemoteException ("Unable to add web application instance to group.");
       }
       int maxClusterSize = (appProperties.getProperty (Context.CLUSTERING) != null) ?
           Integer.parseInt (appProperties.getProperty (Context.CLUSTERING)) : 0;
       if (maxClusterSize > 0) {
         // Start clusteringThread (will form the application cluster, and keep it alive)
         // Check for correct cluster size each minute
         Timer t = new Timer();
         t.scheduleAtFixedRate (new ClusteringTimer (groupName, maxClusterSize, maxClusterSize, false), 20000, 60000);
       }
     }

     deploySnapWebServices("http://" + serverName + ":" + serverPort + "/" + appProperties.getProperty (Context.SNAP_APPCONTEXT),appProperties);
  }

  public DHTHandler getStorageContext() {
    return storageContext;
  }

  public BittorrentHandler getConnectedBTHandler() {
    if (!btHandler.isConnected()) btHandler.connect();
	return btHandler;
  }


  //deploy SNAP p2pWeb Services (WSDD)	  
	public void deploySnapWebServices(String appURL, Properties appProperties) 
	{
	
		Collection params = new Vector();
		String urlAdminService = "-l"+appURL+"/services/AdminService";

		System.out.println("Reading wsdd files....");
		int i = 1;
		while (appProperties.getProperty (Context.SNAP_P2PWS + i) != null) 
		{

			String wsdd_file = (String) appProperties.getProperty (Context.SNAP_P2PWS + i);
			System.out.println(wsdd_file);		  	  
			String workDirBase = appProperties.getProperty (Context.SNAP_CODEBASE);
			String wsdd_path = workDirBase + File.separator + "WEB-INF" + File.separator + wsdd_file;
			System.out.println(wsdd_path);
	  
			params.add(wsdd_path);
			//deploy webservice on work directory
			i++;
		}

		if (!params.isEmpty()) new Thread(new WSDDThread(params, urlAdminService)).start();
	}

	class WSDDThread extends Thread 
	{
		private Collection params;
		private String urlAdminService;

		public WSDDThread(Collection params, String urlAdminService) 
		{
			this.params = params;
			this.urlAdminService = urlAdminService;
		}

		public void run() 
		{
			AdminClient admin = new AdminClient();			
			Iterator it = params.iterator();
			int i = 1;

			while(it.hasNext()) 
			{
				try  
				{						
					System.out.println("Deploying wsdd num "+i+ "...");
					String wsdd_path = (String) it.next();
					admin.process(new String[]{urlAdminService,wsdd_path});		
				} 
				catch (Exception e) 
				{
					System.out.println("Error in deploy process of wsdd num "+i+ "...");			
				}
				i++;
			}			
		}

	}

  /*
  public void deploySnapWebServices(String appURL, Properties appProperties) {
	
    System.out.println("Reading wsdd files....");
    int i = 1;
    while (appProperties.getProperty (Context.SNAP_P2PWS + i) != null) {

      String wsdd_file = (String) appProperties.getProperty (Context.SNAP_P2PWS + i);
	  System.out.println(wsdd_file);		  	  
	  String workDirBase = appProperties.getProperty (Context.SNAP_CODEBASE);
	  String wsdd_path = workDirBase + File.separator + "WEB-INF" + File.separator + wsdd_file;
	  System.out.println(wsdd_path);
	  
	  //deploy webservice on work directory
	  System.out.println("Deploying wsdd num "+i+ "...");
	  try  {
		AdminClient admin = new AdminClient();
		  //-lhttp://sonne:8080/dcalendar/services/AdminService
		admin.process(new String[]{"-l"+appURL+"/services/AdminService",wsdd_path});		
	  } catch (Exception e) {
	    System.out.println("Error in deploy process of wsdd num "+i+ "...");			
	  }
      i++;
    }
   }
   */


  public String getServerName() 
  {
    return serverName;
  }

  public int getServerPort() {
    return serverPort;
  }

  public void setServerPort (int port) {
    this.serverPort = port;

    // Write server port info (for other apps)
    try {
      DataOutputStream d = new DataOutputStream (new FileOutputStream ("./.keepme"));
      d.writeInt (port);
    } catch (Exception e) { e.printStackTrace(); }
  }

  public String getP2PUrl() {
    return appProperties.getProperty (Context.SNAP_APPURL);
  }

  public String getName() {
    return appProperties.getProperty (Context.SNAP_APPNAME);
  }

  public String getContext() {
    return appProperties.getProperty (Context.SNAP_APPCONTEXT);
  }

  public String getWarName() {
    return appProperties.getProperty (Context.SNAP_WARFILE);
  }

  public synchronized void setLastDatabaseCall (long lastCall) {
    lastDatabaseCall = lastCall;
  }

  public String getAppInstance (String appP2PUrl) throws RemoteException {
    return ("http://" + serverName + ":" + serverPort + "/" + appProperties.getProperty (Context.SNAP_APPCONTEXT));
  }

  public boolean getAppInstanceCondition (String appP2PUrl) throws RemoteException {
    return true;
    // Condition is attained if web application signature verifies
/*    Properties appMetadata = (Properties) Registry.lookupSer (appP2PUrl + "_METADATA");

    // Obtain application's signature
    byte[] sig = ((ByteWrapper) appMetadata.get (Context.SNAP_APPSIGNATURE)).getBytes();
    byte[] existentFileBytes = loadFile ("./webapps" + File.separator + appMetadata.get (Context.SNAP_WARFILE));

    if (existentFileBytes == null) {
      System.out.println ("Non-existent WAR file for an active instance: this is very very bad...");
      return false;
    }

    try {
      if (verifySignature (sig, existentFileBytes)) {
        return true;
      }
      else return false;
    } catch (Exception e) {
      return false;
    } */
  }

  public dermi.core.NodeHandle getAppLocalNodeHandle() {
    return conn.getLocalNodeHandle();
  }

  public static Properties getSnapAppConfig (String appRoot) throws ApplicationMetadataIncompleteException {
    Properties snap_war = new Properties();
    Properties config = null;

    try {
      snap_war.loadFromXML (new FileInputStream (appRoot + File.separator + "meta-inf" + File.separator + "snap-war.xml"));
      config = Registry.getEnvironment (appRoot + File.separator + "WEB-INF" + File.separator + "dermi-config.xml");
    } catch (Exception e) {
      e.printStackTrace();
      throw new ApplicationMetadataIncompleteException ("Unable to get SNAP application (" + appRoot + ") configuration.");
    }
    config.putAll (snap_war);
    config.setProperty (dermi.util.Context.CLASS_NAME, Application.class.getName());
    config.setProperty (dermi.util.Context.CLASS_ID, config.getProperty (Context.SNAP_APPURL));
    config.setProperty (Context.BUNSHIN_PROPS, appRoot + File.separator + "WEB-INF" + File.separator + "bunshin.properties");
    config.setProperty (Context.SNAP_CODEBASE, appRoot);

    return config;
  }

  public static void bindSnapRoot() throws RemoteException {
    try {
      Registry.lookupSer ("p2p://snap.deim.urv.net");
    } catch (NotBoundException ne) {
      System.out.println ("Base Snap root not bound... Binding 'p2p://snap.deim.urv.net'");
      Hashtable p = new Hashtable();
      p.put (dermi.util.Context.DERMI_DOLR_UNIQUE_URI, "p2p://snap.deim.urv.net");
      p.put (dermi.util.Context.DERMI_DOLR_CHILDREN, new ArrayList<Properties>());
      Registry.bind ("p2p://snap.deim.urv.net", p);
    }
  }
  private byte[] loadFile (String fileName) {
    try {
      File f = new File (fileName);

      // Load and check if file contains the administrator's signature untampered
      InputStream is = new BufferedInputStream (new FileInputStream (f));
      int length = (int) f.length ();
      byte[] existentFileBytes = new byte[length];

      if (length > Integer.MAX_VALUE) {
        // File is too large
        throw new ApplicationDeploymentException ("Application file to read (" + f.getAbsolutePath () + ") is too large!");
      }

      // Read in the bytes
      int offset = 0;
      int numRead = 0;
      while (offset < existentFileBytes.length &&
             (numRead = is.read (existentFileBytes, offset,
                                 existentFileBytes.length - offset)) >= 0) {
        offset += numRead;
      }

      // Ensure all the bytes have been read in
      if (offset < existentFileBytes.length) {
        throw new ApplicationDeploymentException ("Could not completely read file " + f.getAbsolutePath ());
      }

      // Close the input stream
      is.close ();
      return existentFileBytes;
    } catch (Exception e) {
    }
    return null;
  }

  public Properties redirectToSnapApp (String p2pUrl, String webAppDir, IApplication snapSkel) throws ApplicationDeploymentException, RemoteException {
    // Anycall application's name group (p2p://...) and get closest node's IP+port (if available)
    // Check if an available war file already exists (and redirect to it if signature verifies)
    try {
      Hashtable paux = (Hashtable) Registry.lookupSer (p2pUrl + "_METADATA");
		Properties appMetadata = new Properties();
		appMetadata.putAll(paux);

      // Obtain application's signature
      byte[] sig = ((ByteWrapper) appMetadata.get (Context.SNAP_APPSIGNATURE)).getBytes ();
      byte[] existentFileBytes = loadFile (webAppDir + File.separator + appMetadata.get (Context.SNAP_WARFILE));

      // If file exists, verify signature
      if (existentFileBytes != null) {
        if (verifySignature (sig, existentFileBytes)) {
          logger.info ("File to deploy: " +
                       appMetadata.getProperty (Context.SNAP_WARFILE) +
                       " is already deployed, and signature matches. Do not overwrite.");
          Properties p = new Properties ();
          p.setProperty ("LOCAL", "true");
          p.setProperty ("URL",
                         appMetadata.getProperty (Context.SNAP_APPCONTEXT));
          p.setProperty ("TIMEOUT", "false");
          System.out.println ("WAR file signature verifies.");
          return p;
        }
        else {
          System.out.println ("Signature does not verify!");
        }
      }

      String ip = snapSkel.getAppInstance (p2pUrl);
      Properties p = new Properties();
      p.setProperty ("LOCAL", "false");
      p.setProperty ("URL", ip);
      p.setProperty ("TIMEOUT", "false");
      System.out.println ("Already running webapp instance found... Redirecting to it.");
      return p;
    } catch (Exception remex) {
      System.out.println ("No other running webapp instance found... Deploying locally.");
      return deploySnapApp (p2pUrl, webAppDir);
    }
  }

  private boolean verifySignature (byte[] signature, byte[] fileBytes) throws RemoteException, VerificationException {
    try {
      // Check if the application's signature is validated against the administrator's public key
      Signature rsa = Signature.getInstance ("SHA1withRSA");

      // Initialize the object with the administrator's public key
      PublicKey pubKey = (PublicKey) Registry.lookupSer (Context.SNAP_ADMIN_PK);
      rsa.initVerify (pubKey);

      // Update and verify the data
      rsa.update (fileBytes);
      return rsa.verify (signature);
    } catch (java.security.NoSuchAlgorithmException ne) {
      throw new VerificationException ("No such algorithm: SHA1withRSA!");
    } catch (java.security.InvalidKeyException ik) {
      throw new VerificationException ("Invalid administrator's public key!");
    } catch (java.security.SignatureException se) {
      throw new VerificationException ("Signature exception!");
    }
  }

  
  private Properties deploySnapApp (String p2pUrl, String webAppDir) throws ApplicationDeploymentException, RemoteException {

	  Hashtable paux = (Hashtable) Registry.lookupSer (p2pUrl + "_METADATA");
	  Properties appMetadata = new Properties();
	  appMetadata.putAll(paux);

    // Obtain application's signature
    byte[] sig = ((ByteWrapper) appMetadata.get (Context.SNAP_APPSIGNATURE)).getBytes();

    // Check whether file already exists
    byte[] existentFileBytes = loadFile ((webAppDir + File.separator + appMetadata.getProperty (Context.SNAP_WARFILE)));

    try {
      if (existentFileBytes != null) {
        try {
          if (!verifySignature (sig, existentFileBytes)) {
            System.out.println ("Signature for existent WAR file does not match. Trying to obtain the Registry's version.");
          }
          else {
            logger.info ("File to deploy: " +
                         appMetadata.getProperty (Context.SNAP_WARFILE) +
                         " is already deployed, and signature matches. Do not overwrite.");
            Properties p = new Properties ();
            p.setProperty ("LOCAL", "true");
            p.setProperty ("URL", appMetadata.getProperty (Context.SNAP_APPCONTEXT));
            p.setProperty ("TIMEOUT", "false");
            return p;
          }
        }
        catch (Exception e) {
          e.printStackTrace ();
          System.out.println ("Unable to verify signature for already deployed WAR file. Proceeding to reload from registry.");
        }
      }

      // Application was not previously deployed anywhere, let's deploy it here ;-)
      ByteWrapper bw = (ByteWrapper) Registry.lookupSer (p2pUrl + "_WARFILE");
      byte[] registryFileBytes = bw.getBytes();

      try {
        if (!verifySignature (sig, registryFileBytes)) {
          throw new ApplicationDeploymentException (
              "Web application incorrectly signed. (" + registryFileBytes.length + " bytes)");
        }
      }
      catch (Exception e) {
        e.printStackTrace();
        throw new RemoteException ("Web application '" + p2pUrl +
                                   " has not been signed by the administrator. Cause " + e.getCause ());
      }

      try {
        File f = new File (webAppDir + File.separator + appMetadata.getProperty (Context.SNAP_WARFILE));
        f.delete();
        FileOutputStream fos = new FileOutputStream (webAppDir + File.separator +
                                                     appMetadata.getProperty (Context.SNAP_WARFILE));
        fos.write (registryFileBytes);
        fos.close();

        // Wait five seconds to complete deployment
        // Thread.currentThread().sleep (5000);

        Properties p = new Properties();
        p.setProperty ("LOCAL", "true");
        p.setProperty ("URL", appMetadata.getProperty (Context.SNAP_APPCONTEXT));
        p.setProperty ("TIMEOUT", "true");
        return p;
      }
      catch (Exception e) {
        e.printStackTrace ();
        throw new RemoteException (
            "Unable to create WAR file on webserver's web application directory...!");
      }
    } catch (Exception e) {
      e.printStackTrace();
      if (e instanceof RemoteException) {
        throw (RemoteException) e;
      }
      if (e instanceof ApplicationDeploymentException) {
        throw (ApplicationDeploymentException) e;
      }
      throw new RemoteException (e);
    }
  }

  public static IApplication loadMainSNAPApp (String appRoot, String p2pUrl) throws RemoteException {
    try {
      Properties p = getSnapAppConfig (appRoot);
      p.setProperty (dermi.util.Context.CLASS_NAME, Application.class.getName());
      p.setProperty (dermi.util.Context.CLASS_ID, p2pUrl);
      return (IApplication) dermi.registry.FileRegistry.loadERef (p);
    } catch (Exception e) {
      e.printStackTrace();
      throw new RemoteException (e);
    }
  }

  private void instantiateDBServer (String appContext, String groupName, int membersNum, int maxClusterSize) throws RemoteException {
    int dbPort = firstUsableDBPort;

    while (usedDBPorts.contains (new Integer (dbPort))) {
      dbPort += 2;
    }

    do {
      dbException = false;
      System.setProperty ("group", groupName);
      s = new org.hsqldb.Server();
      s.setPort (dbPort);
      String serverProps = "database.0=file:" +
          appProperties.getProperty (Context.SNAP_CODEBASE) +
          appProperties.getProperty (Context.SNAP_DBDATAPATH) +
          appContext;
      s.putPropertiesFromString (serverProps);
      s.start ();

      int timeout = 0;
      while (timeout < 5000 && s.getStateDescriptor().equals ("SHUTDOWN")) {
        try {
          Thread.currentThread ().sleep (50);
        } catch (InterruptedException e) {}
        timeout += 50;
        // Check if an exception occured
        if (s.getServerError () != null) {
          dbException = true;
          System.out.println ("Unable to bind HSQLDB server on port " + dbPort + ". Trying " + (dbPort + 2));
          break;
        }
      }

      if (!dbException) {
        usedDBPorts.add (new Integer (dbPort));
        lastDatabaseCall = System.currentTimeMillis();
        databaseBound = true;
        addWebAppMember();  // Update membership status: database is bound
        System.out.println ("HSQLDB successfully bound on port " + dbPort);
      }
      else {
        dbPort += 2;
      }
    }
    while (dbException);

    // Let some time for group joining to be complete
//    Thread.currentThread().sleep (5000);

    if (maxClusterSize > 0) {
        // Start clusteringThread (will form the database cluster, and keep it alive)
        // Check for corret cluster size each minute
        Timer t = new Timer();
        t.scheduleAtFixedRate (new ClusteringTimer (groupName, ++membersNum, maxClusterSize, true), 20000, 60000);
      }
  }

  public Connection getConnection() throws SQLException {
    return getConnection ("sa", "");
  }

  /**
   * Start database if not started, and return a connection
   * @return Connection
   * @throws SQLException
   */
  public Connection getConnection (String username, String passwd) throws SQLException {
    if ((appProperties.getProperty (Context.SNAP_PERSISTENCE) != null) &&
        (appProperties.getProperty (Context.SNAP_PERSISTENCE).equals (Context.DATABASE))) {
      try {
        // Load HSQLDB's jdbcDriver
        Class.forName ("org.hsqldb.jdbcDriver");

        if (!databaseBound) {
          String groupName = appProperties.getProperty (Context.SNAP_APPURL);

          int membersNum = getDBMembers().size() - 1;
          int maxClusterSize = (appProperties.getProperty (Context.CLUSTERING) != null) ?
              Integer.parseInt (appProperties.getProperty (Context.CLUSTERING)) : 0;

          // If no 'clustering' (or ==0) property on snap-war.xml, default policy: activate a new database
          //                                      replica in every new node without clustering maintenance
          // If membersNum < 'clustering' -> need to start at least a new db server locally
          if ((maxClusterSize == 0) || (membersNum < maxClusterSize)) {
            // Instantiate new server
            instantiateDBServer (appProperties.getProperty (Context.SNAP_APPCONTEXT), groupName, membersNum, maxClusterSize);
          }
        }
      } catch (ClassNotFoundException e) {
        throw new SQLException ("Driver class not found exception.");
      } catch (RemoteException re) {
        throw new SQLException ("Remote exception connecting to the membership service.");
      }

      // Dynamic proxy for intercepting database activity
      return (Connection) DBInterceptor.newInstance (DriverManager.getConnection ("jdbc:hsqldb:hsql://localhost:" + (usedDBPorts.get (usedDBPorts.size() - 1)), username, passwd), this);
    }
    else {
      throw new SQLException ("Application's persistence mode is not database.");
    }
  }

  /**
   * Start database if not started, and return a connection
   * @return Connection
   * @throws SQLException
   */
  public int dataSourceCreateDB() throws SQLException {
    if ((appProperties.getProperty (Context.SNAP_PERSISTENCE) != null) &&
        (appProperties.getProperty (Context.SNAP_PERSISTENCE).equals (Context.DATABASE))) {
      try {
        if (!databaseBound) {
          String groupName = appProperties.getProperty (Context.SNAP_APPURL);

          int membersNum = getDBMembers().size() - 1;
          int maxClusterSize = (appProperties.getProperty (Context.CLUSTERING) != null) ?
              Integer.parseInt (appProperties.getProperty (Context.CLUSTERING)) : 0;

          // If no 'clustering' (or ==0) property on snap-war.xml, default policy: activate a new database
          //                                      replica in every new node without clustering maintenance
          // If membersNum < 'clustering' -> need to start at least a new db server locally
          if ((maxClusterSize == 0) || (membersNum < maxClusterSize)) {
            // Instantiate new server
            instantiateDBServer (appProperties.getProperty (Context.SNAP_APPCONTEXT), groupName, membersNum, maxClusterSize);
          }
        }
        return (Integer) usedDBPorts.get (usedDBPorts.size() - 1);
      } catch (RemoteException re) {
        throw new SQLException ("Remote exception connecting to the membership service.");
      }
    }
    else {
      throw new SQLException ("Application's persistence mode is not database.");
    }
  }

  private Collection<dermi.core.NodeHandle> getAppMembers() throws RemoteException {
    String group = appProperties.getProperty (Context.SNAP_APPURL);

    // Registry is already loaded: configFile parameter can be null
    if (membership == null) {
      try {
        memberFactory = new MembershipFactory (ComponentUtil.getComponentProperties (MembershipFactory.class, Membership.class, MembershipImpl.class), null);
        membership = (Membership) memberFactory.createInstance ("org.hsqldb.replication.Membership", group);
System.out.println ("CREAT AIXO = " + membership.getERef());
      } catch (ComponentMetadataIncompleteException cmi) {
        cmi.printStackTrace();
      } catch (InterfaceNotFoundException inf) {
        inf.printStackTrace();
      }
    }

    Collection<MembershipData> mbrs = null;
    try {
System.out.println ("BUSCO AIXO = " + membership.getERef());
      mbrs = membership.getMembers();
      System.out.println ("getAppMembers(): members of group " + group + " = " + mbrs.size ());
    } catch (RemoteException re) {
      if (re instanceof dermi.exception.TimeoutException) {
        System.out.println ("TimeoutException: re-creating membership instance");
        try {
          membership = (Membership) memberFactory.createInstance ("org.hsqldb.replication.Membership", group);
System.out.println ("TORNAT A CREAR COMP AMB PROPS = " + membership.getERef());
        } catch (InterfaceNotFoundException inf) {
          inf.printStackTrace();
        }
        mbrs = membership.getMembers();
        System.out.println ("getAppMembers(): members of group " + group + " = " + mbrs.size ());
      }
    }

    ArrayList<dermi.core.NodeHandle> a = new ArrayList<dermi.core.NodeHandle>();
    Iterator<MembershipData> itt = mbrs.iterator();

    while (itt.hasNext()) {
      a.add (itt.next().getNodeHandle());
    }

    return a;
  }

  /**
   * This method returns all members which have a database instance active in this moment
   * @param p2pUrl String Webapp url
   * @return Collection<String> IP+port node listing
   */
  public Collection<dermi.core.NodeHandle> getDBMembers() throws RemoteException {
    String group = appProperties.getProperty (Context.SNAP_APPURL);

    // Registry is already loaded: configFile parameter can be null
    if (membership == null) {
      try {
        memberFactory = new MembershipFactory (ComponentUtil.getComponentProperties (MembershipFactory.class, Membership.class, MembershipImpl.class), null);
        membership = (Membership) memberFactory.createInstance ("org.hsqldb.replication.Membership", group);
      } catch (ComponentMetadataIncompleteException cmi) {
        cmi.printStackTrace();
      } catch (InterfaceNotFoundException inf) {
        inf.printStackTrace();
      }
    }

    Collection<MembershipData> mbrs = null;
    try {
      mbrs = membership.getMembers();
    } catch (RemoteException re) {
      if (re instanceof dermi.exception.TimeoutException) {
        System.out.println ("TimeoutException: re-creating membership instance");
        try {
          membership = (Membership) memberFactory.createInstance ("org.hsqldb.replication.Membership", group);
        } catch (InterfaceNotFoundException inf) {
          inf.printStackTrace();
        }
        mbrs = membership.getMembers();
      }
    }

    ArrayList<dermi.core.NodeHandle> a = new ArrayList<dermi.core.NodeHandle>();
    Iterator<MembershipData> itt = mbrs.iterator();

    while (itt.hasNext()) {
      MembershipData mem = itt.next();
      if (mem.isDatabaseActive()) {
        a.add (mem.getNodeHandle());
      }
    }

    return a;
  }

  private void addWebAppMember() throws RemoteException {
    System.out.println ("addWebAppMember(): connecting...");
    String group = appProperties.getProperty (Context.SNAP_APPURL);

    // Registry is already loaded: configFile parameter can be null
    if (membership == null) {
      try {
        memberFactory = new MembershipFactory (ComponentUtil.getComponentProperties (MembershipFactory.class, Membership.class, MembershipImpl.class), null);
        membership = (Membership) memberFactory.createInstance ("org.hsqldb.replication.Membership", group);
      } catch (ComponentMetadataIncompleteException cmi) {
        cmi.printStackTrace();
      } catch (InterfaceNotFoundException inf) {
        inf.printStackTrace();
      }
    }

    try {
      membership.join (getAppLocalNodeHandle(), databaseBound);
      System.out.println ("addWebAppMember(): added new member to " + group + ": done.");
    } catch (RemoteException re) {
      if (re instanceof dermi.exception.TimeoutException) {
        System.out.println ("TimeoutException: re-creating membership instance");
        membership.join (getAppLocalNodeHandle(), databaseBound);
        System.out.println ("addWebAppMember(): added new member to " + group + ": done.");
      }
    }
  }

  /**
   * This is a remotely-invoked method for deploying a new web application
   * @param nh NodeHandle Since it is a direct call, the first parameter is the destination's nodehandle
   * @throws RemoteException If database cannot be instantiated
   */
  public Object deploySnapAppReplica (dermi.core.NodeHandle nh, String p2pUrl, String webAppDir) throws ApplicationDeploymentException, RemoteException {
    deploySnapApp (p2pUrl, webAppDir);
    try {
      return context.lookup (p2pUrl.replace ('/', '_'));
    } catch (NameNotFoundException nnf) {
      // TODO: Add max timeout here!
      for (;;) {
        try {
          Thread.currentThread().sleep (500);
          return context.lookup (p2pUrl.replace ('/', '_'));
        } catch (NameNotFoundException nef) {
        } catch (Exception e) {
          e.printStackTrace();
          return null;
        }
      }
    } catch (NamingException e) {
      e.printStackTrace();
      throw new RemoteException ("Unable to access JNDI tree.");
    }
  }

  /**
   * This is a remotely-invoked method for instantiating a new database instance on the local node
   * @param nh NodeHandle Since it is a direct call, the first parameter is the destination's nodehandle
   * @param dbGroup String Database groupName
   * @param clusterSize int Desired cluster size
   * @throws RemoteException If database cannot be instantiated
   */
  public void startNewDatabaseInstance (dermi.core.NodeHandle nh, String dbGroup, int clusterSize) throws RemoteException {
    try {
      // Instantiate new DB server, specifying that desired cluster size has been reached (in order to
      // avoid undesirable effects (such as database rebinding on some nodes))
      String dbName = dbGroup.replace ('/', '_');
      dbName = dbName.replace (':', '_');
      instantiateDBServer ("/otherDBs/" + dbName, dbGroup, clusterSize, clusterSize);
    } catch (Exception e) {
      e.printStackTrace();
      throw new RemoteException (e.getMessage());
    }
  }

  /**
   * REMOTE METHOD!
   * @param nh NodeHandle
   * @return boolean
   * @throws RemoteException
   */
  public boolean isDatabaseAlive (dermi.core.NodeHandle nh) throws RemoteException {
    return databaseBound;
  }

  private class ClusteringTimer extends TimerTask {
    private int currentMembersNum;
    private int desiredMembersNum;
    private String groupName;
    private boolean useDatabase;
    private IApplication mainSnapStub;
    private IApplication appStub;

    public ClusteringTimer (String group, int curMembers, int desMembers, boolean uDb) {
      currentMembersNum = curMembers;
      desiredMembersNum = desMembers;
      groupName = group;
      useDatabase = uDb;

      try {
        Properties p = (Properties) appProperties.clone();
        p.setProperty (dermi.util.Context.CLASS_NAME, "org.planet.snap.Application");
        p.setProperty (dermi.util.Context.CLASS_ID, group);

        appStub = (IApplication) FileRegistry.loadERef (p);
        p.setProperty (dermi.util.Context.CLASS_ID, "p2p://snap.urv.net");
        mainSnapStub = (IApplication) FileRegistry.loadERef (p);
      } catch (RemoteException re) {
        re.printStackTrace();
        System.out.println ("Unable to get application stubs.");
      }
    }

    public void run() {
      // Discover actual group members
      try {
          // Check if database is to be passivated ?
          if (lastDatabaseCall + (dbPassivationThreshold * 60 * 1000) < System.currentTimeMillis() && useDatabase) {
            // Passivation time
            logger.info ("Database passivation time. Not received requests for " + dbPassivationThreshold + " minutes.");
            logger.info ("lastCall = " + lastDatabaseCall + " threshold = " + dbPassivationThreshold * 60 * 100 + " currentTime = " + System.currentTimeMillis());
            databaseBound = false;
            addWebAppMember(); // Update membership status: database is not bound
            usedDBPorts.remove (usedDBPorts.size() - 1);
            this.cancel();

            // Close database server
            s.shutdown();
          }
          else if (databaseBound) {
            Collection<dermi.core.NodeHandle> mbrs = getDBMembers();
            currentMembersNum = mbrs.size();
            int diff = desiredMembersNum - currentMembersNum;

            if (diff > 0) {
              // Activate new HSQLDBs in other surrounding nodes
              System.out.println ("Rebalancing cluster... - (currentMembers = " + currentMembersNum + " desiredMembers = " + desiredMembersNum + ")");;
              Collection neighbours = conn.getReplicaSet (conn.getLocalNodeId(), desiredMembersNum * 10);

              ArrayList<rice.p2p.commonapi.NodeHandle> usableNeighbours = new ArrayList();
              Iterator i = neighbours.iterator();

               while (i.hasNext()) {
                boolean found = false;
                // This is ***** VERY UGLY ***** but necessary right now to get NodeHandle's IP Address
                rice.p2p.commonapi.NodeHandle nh = (rice.p2p.commonapi.NodeHandle) i.next();
                // Check if this neighbour is already a member of the database group, and if so,
                // we cannot use it (replication takes place in different physical nodes)
                Iterator it = mbrs.iterator();
                while (it.hasNext()) {
                  // This is ***** VERY UGLY ***** but necessary right now to get NodeHandle's IP Address
                  dermi.core.NodeHandle nodeHandle = (dermi.core.NodeHandle) it.next();
                  String address1 = ((rice.pastry.dist.DistNodeHandle) nodeHandle).getAddress().getAddress().getHostAddress();
                  String address2 = ((rice.pastry.dist.DistNodeHandle) nh).getAddress().getAddress().getHostAddress();

                  if (address1.startsWith (address2)) {
                    found = true;
                  }
                }
                if (!found) {
                  usableNeighbours.add (nh);
                }
              }

           // usableNeighbours = new ArrayList (neighbours);

           if (usableNeighbours.size() > 0) {
             // Activate replicas in all usableNeighbours
             Iterator ite = usableNeighbours.iterator();
             int repz = 0;

             while (ite.hasNext() && repz < diff) {
               dermi.core.NodeHandle nh = (dermi.core.NodeHandle) ite.next();
               if (!nh.equals (conn.getLocalNodeHandle()) && ((rice.p2p.commonapi.NodeHandle) nh).isAlive()) {

                 System.out.println ("Deploying new replica on: " + nh);

                 dermi.core.NodeHandle nnh = (dermi.core.NodeHandle) mainSnapStub.deploySnapAppReplica (nh, appProperties.getProperty (Context.SNAP_APPURL), "./webapps");
                 if (useDatabase) {
                   appStub.startNewDatabaseInstance (nnh, groupName, desiredMembersNum);
                 }
                 repz++;
               }
             }
           } else {
             System.out.println ("Not enough neighbours. Maintaining actual cluster size.");
           }
         }
         else {
           System.out.println ("Database cluster is alive and balanced - (currentMembers = " + currentMembersNum + " desiredMembers = " + desiredMembersNum + ")");
         }
       }
     } catch (Exception e) {
       System.out.println ("Exception occured while trying to balance the DB cluster.");
       e.printStackTrace();
     }
   }
 }

 //////////////////////////////
 // Introspection API
 //////////////////////////////

 /**
  * This method returns a listing of all available Snap web applications
  * which are already deployed.
  * @return String[] List of Snap webapp identifiers.
  */
 public static Collection<String> getApplications() throws RemoteException {
   Collection c = Registry.list ("p2p://snap.deim.urv.net");
   Iterator i = c.iterator();
   ArrayList<String> as = new ArrayList<String>();

   while (i.hasNext()) {
     String s = ((RegistryRecord) i.next()).getObjectURI();
     StringTokenizer tokens = new StringTokenizer (s, "_");
     as.add (tokens.nextToken());
   }

   return as;
 }

 /**
  * This method returns a stub which models a gateway to the specified application
  * @param p2pUrl String Webapp identifier
  * @return ERemote Stub to the Snap webapp
  */
 public static ERemote getApplication (String p2pUrl) throws RemoteException {
   Hashtable env = (Hashtable) Registry.lookupSer (p2pUrl + "_METADATA");
   Properties p = new Properties();
   p.putAll(env);
   p.setProperty (dermi.util.Context.CLASS_NAME, Application.class.getName());
   p.setProperty (dermi.util.Context.CLASS_ID, p2pUrl);   
   return (IApplication) dermi.registry.FileRegistry.loadERef (p);
 }

 private Collection<MembershipData> getInstancesAux() throws RemoteException {
   String group = appProperties.getProperty (Context.SNAP_APPURL);
   // System.out.println ("Getting Snap app instances (" + group + ")");
   Collection<MembershipData> nhs = null;

   // Registry is already loaded: configFile parameter can be null
   if (membership == null) {
     try {
       memberFactory = new MembershipFactory (ComponentUtil.getComponentProperties (MembershipFactory.class, Membership.class, MembershipImpl.class), null);
       membership = (Membership) memberFactory.createInstance ("org.hsqldb.replication.Membership", group);
     } catch (ComponentMetadataIncompleteException ce) {
       ce.printStackTrace();
     } catch (InterfaceNotFoundException inf) {
       inf.printStackTrace();
     }
   }

   try {
     nhs = membership.getMembers();
   } catch (RemoteException re) {
     if (re instanceof dermi.exception.TimeoutException) {
       System.out.println ("TimeoutException: re-creating membership instance");
       nhs = membership.getMembers();
     }
   }
   return nhs;
 }

 /**
  * This method returns all IP+port listing of physical nodes in which this
  * Snap webapp is currently running (returns the nodes of the webapp cluster)
  * @return Collection<String> IP+port node listing
  */
 public Collection<String> getInstances() throws RemoteException {
   Collection<MembershipData> nhs = getInstancesAux();
   ArrayList<String> instances = new ArrayList<String>();
   Iterator<MembershipData> i = nhs.iterator();
   while (i.hasNext()) {
    rice.pastry.dist.DistNodeHandle nh = (rice.pastry.dist.DistNodeHandle) (i.next().getNodeHandle());
    instances.add (nh.getAddress().getHostName() + ":" + nh.getAddress().getPort());
   }
   return instances;
 }

 private void printCollection (Collection<dermi.core.NodeHandle> c) {
   Iterator<dermi.core.NodeHandle> i = c.iterator();
   while (i.hasNext()) {
     rice.pastry.dist.DistNodeHandle nh = (rice.pastry.dist.DistNodeHandle) i.next();
     System.out.println (nh.getAddress().getHostName() + ":" + nh.getAddress().getPort());
   }
 }

 public Collection<String> getDatabaseInstances() throws RemoteException {
   Collection nhs = getDBMembers();
   ArrayList<String> instances = new ArrayList<String>();
   Iterator<dermi.core.NodeHandle> i = nhs.iterator();
   while (i.hasNext()) {
    rice.pastry.dist.DistNodeHandle nh = (rice.pastry.dist.DistNodeHandle) i.next();
    instances.add (nh.getAddress().getHostName() + ":" + nh.getAddress().getPort());
   }
   return instances;
 }

	
}
