package org.planet.snap;

import dermi.*;

import java.io.*;
import java.util.*;
import java.net.*;

import javax.servlet.http.*;
import javax.servlet.*;
import javax.naming.InitialContext;
import javax.naming.NameNotFoundException;

import org.planet.snap.ds.SnapDataSource;

public class SnapAppStartup extends HttpServlet {
  Application snapApp = null;

  public void init (ServletConfig config) throws ServletException {
    super.init (config);
    ServletContext ctx = config.getServletContext();

    try {
        // ----- ATTEMPT TO LOAD LOCAL HOSTNAME -----
        String serverName = InetAddress.getLocalHost().getHostName ();
        ctx.setAttribute ("serverName", serverName);

		InitialContext context = new InitialContext();

        Properties p = Application.getSnapAppConfig (ctx.getRealPath (File.separator));

        // Create main SNAP application handler
        snapApp = new Application (serverName, p);
        ctx.setAttribute ("snapApp", snapApp);
        // Update datasource configuration (if application requires DB persistence)
        if ((p.getProperty (Context.SNAP_PERSISTENCE) != null) &&
            (p.getProperty (Context.SNAP_PERSISTENCE).equals (Context.DATABASE))) {
          try {
            SnapDataSource ds = (SnapDataSource) context.lookup ("java:comp/env/jdbc/SnapDS");
            ds.setApplication (snapApp);
          } catch (NameNotFoundException ne) {
            System.err.println ("Datasource utilization is disabled. If you want to enable DataSources in your webapp, please update your 'web.xml' file.");
          }
        }

        // Rebind application reference
        context.rebind (snapApp.getP2PUrl().replace ('/', '_'), snapApp.getAppLocalNodeHandle());
        System.out.println ("Bound " + snapApp.getP2PUrl().replace ('/', '_') + " to " + snapApp.getAppLocalNodeHandle());

    } catch (Exception e) {
      System.out.println ("------------------------------ EXCEPTION --------------------------------");
      System.out.println ("------------------------ Initializing SNAP App --------------------------");
      System.out.println ("-------------------------------------------------------------------------");
      e.printStackTrace();
      throw new ServletException (e.getMessage());
    }
  }

  public void destroy() {
    super.destroy();

    if (snapApp != null) {
      try {
        Registry.leave ();
        snapApp.close();
      } catch (Exception e) {
        e.printStackTrace();
      }
    }
  }
}

