package org.planet.snap.deployer;

import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.TitleAreaDialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.ProgressBar;

public class KeyGenWizardDialog extends TitleAreaDialog {
  private SnapDeployer win;
  private boolean keyStoreFound;
  private boolean isWar;
  private ProgressBar progressBar;
  private Button next;

  public KeyGenWizardDialog (Shell parentShell, SnapDeployer wd, boolean ksf, boolean isw) {
    super (parentShell);
    win = wd;
    keyStoreFound = ksf;
    isWar = isw;
  }

  protected Control createDialogArea (Composite parent) {
    Composite area = (Composite) super.createDialogArea(parent);
    Composite container = new Composite (area, SWT.NONE);
    container.setLayoutData(new GridData(GridData.FILL_BOTH));

    super.getShell().setText ("Snap Deployer");
    Image img = new Image (super.getShell().getDisplay(), "./images/snap_installer.jpg");
    super.getShell().setImage (img);

    setTitle ("SNAP Web Application Signer and Deployer Wizard");
    if (keyStoreFound) {
      setMessage ("Key store correctly loaded.");
    }
    else {
      setMessage ("Key pair generation process.");
    }
    setTitleImage (img);

    if (keyStoreFound) {
      final Label label = new Label(container, SWT.NONE);
      label.setBounds(15, 27, 450, 117);
      label.setText("Your public/private key pair has been correctly loaded from the default keystore (file named\n'keystore.rsa').\n\n" +
                    "They will be used from now on for web application signature and verification.\n\n\n\nPlease press Next button to continue.");
    }
    else {
      final Label label = new Label(container, SWT.NONE);
      label.setBounds(15, 27, 450, 87);
      label.setText ("No public/private key pair has been found on the default keystore file (file named\r\n'keystore.key').\n\nPlease wait while generating your default public/private key pair.");

      final Label label_1 = new Label(container, SWT.SEPARATOR | SWT.HORIZONTAL);
      label_1.setBounds (7, 200, 485, 5);

      progressBar = new ProgressBar (container, SWT.NONE);
      progressBar.setBounds (56, 119, 369, 10);
      progressBar.setMaximum (100);
      progressBar.setMinimum (0);

      final Button generateButton = new Button(container, SWT.NONE);
      generateButton.setBounds (177, 144, 120, 25);
      generateButton.setText ("Generate!");
      generateButton.addSelectionListener (new SelectionAdapter() {
        public void widgetSelected (SelectionEvent event) {
          try {
            win.generateKeyPair (progressBar);
          } catch (Exception e) {
            e.printStackTrace();
          }
          generateButton.setEnabled (false);
          next.setEnabled (true);
        }
      });
    }
    return area;
  }

  protected void createButtonsForButtonBar (Composite parent) {
    createButton (parent, IDialogConstants.BACK_ID, IDialogConstants.BACK_LABEL, false);
    createButton (parent, IDialogConstants.NEXT_ID, IDialogConstants.NEXT_LABEL, true);
    createButton (parent, IDialogConstants.CANCEL_ID, IDialogConstants.CANCEL_LABEL, false);

    next = getButton (IDialogConstants.NEXT_ID);
    if (!keyStoreFound) {
      next.setEnabled (false);
    }
    next.addSelectionListener (new SelectionAdapter() {
      public void widgetSelected (SelectionEvent event) {
        Point p = getShell().getLocation();
        close();
        if (isWar) {
          win.openWarDeploymentDialog (p);
        }
        else {
          win.openTransformWebappDialog (p);
        }
      }
    });

    Button cancel = getButton (IDialogConstants.CANCEL_ID);
    cancel.addSelectionListener (new SelectionAdapter() {
      public void widgetSelected(SelectionEvent event) {
        close();
      }
    });

    Button back = getButton (IDialogConstants.BACK_ID);
    back.addSelectionListener (new SelectionAdapter() {
      public void widgetSelected(SelectionEvent event) {
        Point p = getShell().getLocation();
        close();
        win.openMainOptionsDialog (p);
      }
    });
  }

  protected Point getInitialSize() {
    return new Point (500, 375);
  }
}
